@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2015 gnome-mpv
+ * Copyright (c) 2014-2016 gnome-mpv
  *
  * This file is part of GNOME MPV.
  *
@@ -17,8 +17,16 @@
  * along with GNOME MPV.  If not, see <http://www.gnu.org/licenses/>.
  */
 
+#include <glib/gi18n.h>
+#include <gdk/gdk.h>
+
+#ifdef GDK_WINDOWING_WAYLAND
+#include <gdk/gdkwayland.h>
+#endif
+
 #include "def.h"
 #include "menu.h"
+#include "application.h"
 #include "playlist_widget.h"
 #include "main_window.h"
 #include "control_box.h"
@@ -26,22 +34,66 @@
 enum
 {
 	PROP_0,
+	PROP_PLAYLIST,
 	PROP_USE_OPENGL,
 	N_PROPERTIES
 };
 
 struct _MainWindowPrivate
 {
+	Playlist *playlist;
+	gboolean playlist_first_toggle;
 	gboolean use_opengl;
+	gint width_offset;
+	gint height_offset;
+	gint resize_target[2];
 };
 
+static void main_window_set_property(	GObject *object,
+					guint property_id,
+					const GValue *value,
+					GParamSpec *pspec );
+static void main_window_get_property(	GObject *object,
+					guint property_id,
+					GValue *value,
+					GParamSpec *pspec );
+static gboolean fs_control_enter_handler(	GtkWidget *widget,
+						GdkEvent *event,
+						gpointer data );
+static gboolean fs_control_leave_handler(	GtkWidget *widget,
+						GdkEvent *event,
+						gpointer data );
+static gboolean motion_notify_handler(GtkWidget *widget, GdkEventMotion *event);
+static void size_allocate_handler(	GtkWidget *widget,
+					GdkRectangle *allocation,
+					gpointer data );
 static void vid_area_init(MainWindow *wnd, gboolean use_opengl);
 static GtkWidget *vid_area_new(gboolean use_opengl);
 static gboolean timeout_handler(gpointer data);
-static gboolean finalize_load_state(gpointer data);
+static gboolean resize_to_target(gpointer data);
 
 G_DEFINE_TYPE_WITH_PRIVATE(MainWindow, main_window, GTK_TYPE_APPLICATION_WINDOW)
 
+static void main_window_constructed(GObject *object)
+{
+	MainWindow *self = MAIN_WINDOW(object);
+
+	self->playlist = playlist_widget_new(self->priv->playlist);
+
+	gtk_widget_show_all(self->playlist);
+	gtk_widget_hide(self->playlist);
+	gtk_widget_set_no_show_all(self->playlist, TRUE);
+
+	gtk_paned_pack2(	GTK_PANED(self->vid_area_paned),
+				self->playlist,
+				FALSE,
+				FALSE );
+
+	vid_area_init(self, self->priv->use_opengl);
+
+	G_OBJECT_CLASS(main_window_parent_class)->constructed(object);
+}
+
 static void main_window_set_property(	GObject *object,
 					guint property_id,
 					const GValue *value,
@@ -49,12 +101,14 @@ static void main_window_set_property(	GObject *object,
 {
 	MainWindow *self = MAIN_WINDOW(object);
 
-	/* PROP_USE_OPENGL can only be set once during construction */
-	if(property_id == PROP_USE_OPENGL)
+	if(property_id == PROP_PLAYLIST)
+	{
+		self->priv->playlist = g_value_get_pointer(value);
+
+	}
+	else if(property_id == PROP_USE_OPENGL)
 	{
 		self->priv->use_opengl = g_value_get_boolean(value);
-
-		vid_area_init(self, self->priv->use_opengl);
 	}
 	else
 	{
@@ -69,7 +123,11 @@ static void main_window_get_property(	GObject *object,
 {
 	MainWindow *self = MAIN_WINDOW(object);
 
-	if(property_id == PROP_USE_OPENGL)
+	if(property_id == PROP_PLAYLIST)
+	{
+		g_value_set_pointer(value, self->priv->playlist);
+	}
+	else if(property_id == PROP_USE_OPENGL)
 	{
 		g_value_set_boolean(value, self->priv->use_opengl);
 	}
@@ -102,15 +160,14 @@ static gboolean motion_notify_handler(GtkWidget *widget, GdkEventMotion *event)
 	MainWindow *wnd = MAIN_WINDOW(widget);
 	GdkCursor *cursor;
 
-	cursor = gdk_cursor_new_from_name(	gdk_display_get_default(),
-						"default" );
-
+	cursor = gdk_cursor_new_from_name(gdk_display_get_default(), "default");
 	gdk_window_set_cursor
 		(gtk_widget_get_window(GTK_WIDGET(wnd->vid_area)), cursor);
 
 	if(wnd->fullscreen)
 	{
-		gtk_widget_show(wnd->control_box);
+		gtk_revealer_set_reveal_child
+			(GTK_REVEALER(wnd->fs_revealer), TRUE);
 	}
 
 	if(wnd->timeout_tag > 0)
@@ -126,29 +183,33 @@ static gboolean motion_notify_handler(GtkWidget *widget, GdkEventMotion *event)
 		->motion_notify_event(widget, event);
 }
 
-static gboolean configure_handler(	GtkWidget *widget,
-					GdkEvent *event,
+static void size_allocate_handler(	GtkWidget *widget,
+					GdkRectangle *allocation,
 					gpointer data )
 {
 	MainWindow *wnd = data;
-	GdkEventConfigure *conf_event = (GdkEventConfigure *)event;
-	guint signal_id = g_signal_lookup("configure-event", MAIN_WINDOW_TYPE);
+	GdkScreen *screen = gdk_screen_get_default();
+	gint screen_width = gdk_screen_get_width(screen);
+	gint screen_height = gdk_screen_get_height(screen);
+	gint width = allocation->width;
+	gint height = allocation->height;
+	gint target_width = wnd->priv->resize_target[0];
+	gint target_height = wnd->priv->resize_target[1];
 
-	if(wnd->init_width == conf_event->width)
+	g_signal_handlers_disconnect_by_func
+		(widget, size_allocate_handler, data);
+
+	/* Adjust resize offset */
+	if((width != target_width || height != target_height)
+	&& (target_width < screen_width && target_height < screen_height)
+	&& !gtk_window_is_maximized(GTK_WINDOW(wnd))
+	&& !wnd->fullscreen)
 	{
-		g_signal_handlers_disconnect_matched(	wnd,
-							G_SIGNAL_MATCH_ID
-							|G_SIGNAL_MATCH_DATA,
-							signal_id,
-							0,
-							0,
-							NULL,
-							data);
+		wnd->priv->width_offset += target_width-width;
+		wnd->priv->height_offset += target_height-height;
 
-		g_idle_add((GSourceFunc)finalize_load_state, data);
+		g_idle_add(resize_to_target, wnd);
 	}
-
-	return FALSE;
 }
 
 static void vid_area_init(MainWindow *wnd, gboolean use_opengl)
@@ -156,6 +217,7 @@ static void vid_area_init(MainWindow *wnd, gboolean use_opengl)
 	/* vid_area cannot be initialized more than once */
 	if(!wnd->vid_area)
 	{
+		GtkTargetEntry targets[] = DND_TARGETS;
 		GtkStyleContext *style_context;
 
 		wnd->vid_area =	vid_area_new(use_opengl);
@@ -163,6 +225,13 @@ static void vid_area_init(MainWindow *wnd, gboolean use_opengl)
 
 		gtk_style_context_add_class(style_context, "gmpv-vid-area");
 
+		gtk_drag_dest_set(	wnd->vid_area,
+					GTK_DEST_DEFAULT_ALL,
+					targets,
+					G_N_ELEMENTS(targets),
+					GDK_ACTION_LINK );
+		gtk_drag_dest_add_uri_targets(wnd->vid_area);
+
 		/* GDK_BUTTON_RELEASE_MASK is needed so that GtkMenuButtons can
 		 * hide their menus when vid_area is clicked.
 		 */
@@ -172,7 +241,6 @@ static void vid_area_init(MainWindow *wnd, gboolean use_opengl)
 
 		gtk_container_add(	GTK_CONTAINER(wnd->vid_area_overlay),
 					wnd->vid_area );
-
 		gtk_paned_pack1(	GTK_PANED(wnd->vid_area_paned),
 					wnd->vid_area_overlay,
 					TRUE,
@@ -182,40 +250,31 @@ static void vid_area_init(MainWindow *wnd, gboolean use_opengl)
 
 static GtkWidget *vid_area_new(gboolean use_opengl)
 {
-#ifdef OPENGL_CB_ENABLED
 	return use_opengl?gtk_gl_area_new():gtk_drawing_area_new();
-#else
-	return gtk_drawing_area_new();
-#endif
 }
 
 static gboolean timeout_handler(gpointer data)
 {
 	MainWindow *wnd;
 	ControlBox *control_box;
-	GtkScaleButton *button;
-	GtkWidget *popup;
 
 	wnd = data;
 	control_box = CONTROL_BOX(wnd->control_box);
-	button = GTK_SCALE_BUTTON(control_box->volume_button);
-	popup = gtk_scale_button_get_popup(button);
-
 
 	if(wnd->fullscreen
 	&& !wnd->fs_control_hover
-	&& !gtk_widget_is_visible(popup))
+	&& !control_box_get_volume_popup_visible(control_box))
 	{
 		GdkWindow *window;
 		GdkCursor *cursor;
 
 		window = gtk_widget_get_window(GTK_WIDGET(wnd->vid_area));
-
 		cursor = gdk_cursor_new_for_display
 				(gdk_display_get_default(), GDK_BLANK_CURSOR);
 
 		gdk_window_set_cursor(window, cursor);
-		gtk_widget_hide(wnd->control_box);
+		gtk_revealer_set_reveal_child
+			(GTK_REVEALER(wnd->fs_revealer), FALSE);
 	}
 
 	wnd->timeout_tag = 0;
@@ -223,24 +282,237 @@ static gboolean timeout_handler(gpointer data)
 	return FALSE;
 }
 
-static gboolean finalize_load_state(gpointer data)
+static gboolean resize_to_target(gpointer data)
 {
 	MainWindow *wnd = data;
+	gint target_width = wnd->priv->resize_target[0];
+	gint target_height = wnd->priv->resize_target[1];
 
-	gtk_paned_set_position(	GTK_PANED(wnd->vid_area_paned),
-				wnd->init_width-wnd->playlist_width );
-
-	if(wnd->init_playlist_visible != main_window_get_playlist_visible(wnd))
-	{
-		wnd->playlist_visible = wnd->init_playlist_visible;
-
-		gtk_widget_set_visible(	wnd->playlist,
-					wnd->init_playlist_visible );
-	}
+	gtk_window_resize(	GTK_WINDOW(wnd),
+				target_width+wnd->priv->width_offset,
+				target_height+wnd->priv->height_offset );
 
 	return FALSE;
 }
 
+static void main_window_class_init(MainWindowClass *klass)
+{
+	GObjectClass *obj_class = G_OBJECT_CLASS(klass);
+	GtkWidgetClass *wgt_class = GTK_WIDGET_CLASS(klass);
+	GParamSpec *pspec = NULL;
+
+	obj_class->constructed = main_window_constructed;
+	obj_class->set_property = main_window_set_property;
+	obj_class->get_property = main_window_get_property;
+	wgt_class->motion_notify_event = motion_notify_handler;
+
+	pspec = g_param_spec_pointer
+		(	"playlist",
+			"Playlist",
+			"Playlist object used to store playlist items",
+			G_PARAM_CONSTRUCT_ONLY|G_PARAM_READWRITE );
+
+	g_object_class_install_property(obj_class, PROP_PLAYLIST, pspec);
+
+	pspec = g_param_spec_boolean
+		(	"use-opengl",
+			"Use OpenGL",
+			"Whether or not to set up video area for opengl-cb",
+			FALSE,
+			G_PARAM_CONSTRUCT_ONLY|G_PARAM_READWRITE );
+
+	g_object_class_install_property(obj_class, PROP_USE_OPENGL, pspec);
+}
+
+static void main_window_init(MainWindow *wnd)
+{
+	/* wnd->vid_area will be initialized when use-opengl property is set */
+	wnd->priv = main_window_get_instance_private(wnd);
+	wnd->fullscreen = FALSE;
+	wnd->playlist_visible = FALSE;
+	wnd->fs_control_hover = FALSE;
+	wnd->pre_fs_playlist_visible = FALSE;
+	wnd->playlist_width = PLAYLIST_DEFAULT_WIDTH;
+	wnd->timeout_tag = 0;
+	wnd->settings = gtk_settings_get_default();
+	wnd->header_bar = gtk_header_bar_new();
+	wnd->open_hdr_btn = NULL;
+	wnd->fullscreen_hdr_btn = NULL;
+	wnd->menu_hdr_btn = NULL;
+	wnd->main_box = gtk_box_new(GTK_ORIENTATION_VERTICAL, 0);
+	wnd->vid_area_paned = gtk_paned_new(GTK_ORIENTATION_HORIZONTAL);
+	wnd->vid_area_overlay = gtk_overlay_new();
+	wnd->control_box = control_box_new();
+	wnd->fs_revealer = gtk_revealer_new();
+
+	wnd->priv->playlist_first_toggle = TRUE;
+	wnd->priv->width_offset = 0;
+	wnd->priv->height_offset = 0;
+
+	gtk_widget_add_events(	wnd->vid_area_overlay,
+				GDK_ENTER_NOTIFY_MASK
+				|GDK_LEAVE_NOTIFY_MASK );
+
+	gtk_header_bar_set_show_close_button(	GTK_HEADER_BAR(wnd->header_bar),
+						TRUE );
+
+	gtk_window_set_title(GTK_WINDOW(wnd), g_get_application_name());
+
+	gtk_paned_set_position(	GTK_PANED(wnd->vid_area_paned),
+				MAIN_WINDOW_DEFAULT_WIDTH
+				-PLAYLIST_DEFAULT_WIDTH );
+
+	gtk_window_set_default_size(	GTK_WINDOW(wnd),
+					MAIN_WINDOW_DEFAULT_WIDTH,
+					MAIN_WINDOW_DEFAULT_HEIGHT );
+
+	gtk_widget_set_vexpand(wnd->fs_revealer, FALSE);
+	gtk_widget_set_hexpand(wnd->fs_revealer, FALSE);
+	gtk_widget_set_halign(wnd->fs_revealer, GTK_ALIGN_CENTER);
+	gtk_widget_set_valign(wnd->fs_revealer, GTK_ALIGN_END);
+	gtk_widget_show(wnd->fs_revealer);
+	gtk_revealer_set_reveal_child(GTK_REVEALER(wnd->fs_revealer), FALSE);
+
+	gtk_overlay_add_overlay
+		(GTK_OVERLAY(wnd->vid_area_overlay), wnd->fs_revealer);
+	gtk_box_pack_start
+		(GTK_BOX(wnd->main_box), wnd->vid_area_paned, TRUE, TRUE, 0);
+	gtk_container_add(GTK_CONTAINER(wnd->main_box), wnd->control_box);
+	gtk_container_add(GTK_CONTAINER(wnd), wnd->main_box);
+
+	g_signal_connect(	wnd->vid_area_overlay,
+				"enter-notify-event",
+				G_CALLBACK(fs_control_enter_handler),
+				wnd );
+	g_signal_connect(	wnd->vid_area_overlay,
+				"leave-notify-event",
+				G_CALLBACK(fs_control_leave_handler),
+				wnd );
+}
+
+GtkWidget *main_window_new(	Application *app,
+				Playlist *playlist,
+				gboolean use_opengl )
+{
+	return GTK_WIDGET(g_object_new(	main_window_get_type(),
+					"application", app,
+					"playlist", playlist,
+					"use-opengl", use_opengl,
+					NULL ));
+}
+
+void main_window_set_fullscreen(MainWindow *wnd, gboolean fullscreen)
+{
+	if(fullscreen != wnd->fullscreen)
+	{
+		ControlBox *control_box = CONTROL_BOX(wnd->control_box);
+		GtkContainer* main_box = GTK_CONTAINER(wnd->main_box);
+		GtkContainer* revealer = GTK_CONTAINER(wnd->fs_revealer);
+
+		if(fullscreen)
+		{
+			GdkScreen *screen;
+			GdkWindow *window;
+			GdkRectangle monitor_geom;
+			gint width;
+			gint monitor;
+
+			screen = gtk_window_get_screen(GTK_WINDOW(wnd));
+			window = gtk_widget_get_window(GTK_WIDGET(wnd));
+			monitor =	gdk_screen_get_monitor_at_window
+					(screen, window);
+
+			gdk_screen_get_monitor_geometry
+				(screen, monitor, &monitor_geom);
+
+			width = monitor_geom.width/2;
+
+			gtk_widget_set_size_request
+				(wnd->control_box, width, -1);
+
+			g_object_ref(wnd->control_box);
+			gtk_container_remove(main_box, wnd->control_box);
+			gtk_container_add(revealer, wnd->control_box);
+			g_object_unref(wnd->control_box);
+
+			control_box_set_fullscreen_state(control_box, TRUE);
+			gtk_window_fullscreen(GTK_WINDOW(wnd));
+			gtk_window_present(GTK_WINDOW(wnd));
+			gtk_widget_show(GTK_WIDGET(revealer));
+
+			if(main_window_get_csd_enabled(wnd))
+			{
+				control_box_set_fullscreen_btn_visible
+					(CONTROL_BOX(wnd->control_box), TRUE);
+			}
+			else
+			{
+				gtk_application_window_set_show_menubar
+					(GTK_APPLICATION_WINDOW(wnd), FALSE);
+			}
+
+			wnd->pre_fs_playlist_visible = wnd->playlist_visible;
+			gtk_widget_set_visible(wnd->playlist, FALSE);
+		}
+		else
+		{
+			GdkCursor *cursor;
+			GdkWindow *vid_area;
+
+			gtk_widget_set_halign(wnd->control_box, GTK_ALIGN_FILL);
+			gtk_widget_set_valign(wnd->control_box, GTK_ALIGN_FILL);
+			gtk_widget_set_size_request(wnd->control_box, -1, -1);
+
+			g_object_ref(wnd->control_box);
+			gtk_container_remove(revealer, wnd->control_box);
+			gtk_container_add(main_box, wnd->control_box);
+			g_object_unref(wnd->control_box);
+
+			control_box_set_fullscreen_state(control_box, FALSE);
+			gtk_window_unfullscreen(GTK_WINDOW(wnd));
+			gtk_widget_hide(GTK_WIDGET(revealer));
+
+			if(main_window_get_csd_enabled(wnd))
+			{
+				control_box_set_fullscreen_btn_visible
+					(CONTROL_BOX(wnd->control_box), FALSE);
+			}
+			else
+			{
+				gtk_application_window_set_show_menubar
+					(GTK_APPLICATION_WINDOW(wnd), TRUE);
+			}
+
+			wnd->playlist_visible = wnd->pre_fs_playlist_visible;
+			gtk_widget_set_visible
+				(wnd->playlist, wnd->pre_fs_playlist_visible);
+
+			cursor =	gdk_cursor_new_from_name
+					(gdk_display_get_default(), "default");
+			vid_area =	gtk_widget_get_window
+					(GTK_WIDGET(wnd->vid_area));
+
+			gdk_window_set_cursor(vid_area, cursor);
+		}
+
+		wnd->fullscreen = fullscreen;
+
+		timeout_handler(wnd);
+	}
+}
+
+void main_window_toggle_fullscreen(MainWindow *wnd)
+{
+	main_window_set_fullscreen(wnd, !wnd->fullscreen);
+}
+
+void main_window_reset(MainWindow *wnd)
+{
+	gtk_window_set_title(GTK_WINDOW(wnd), g_get_application_name());
+	control_box_reset(CONTROL_BOX(wnd->control_box));
+	playlist_set_indicator_pos(PLAYLIST_WIDGET(wnd->playlist)->store, -1);
+}
+
 void main_window_save_state(MainWindow *wnd)
 {
 	GSettings *settings;
@@ -258,10 +530,7 @@ void main_window_save_state(MainWindow *wnd)
 	g_settings_set_int(settings, "width", width);
 	g_settings_set_int(settings, "height", height);
 	g_settings_set_double(settings, "volume", volume);
-
-	g_settings_set_boolean(	settings,
-				"show-playlist",
-				wnd->playlist_visible );
+	g_settings_set_boolean(settings, "show-playlist", wnd->playlist_visible);
 
 	if(main_window_get_playlist_visible(wnd))
 	{
@@ -281,32 +550,34 @@ void main_window_save_state(MainWindow *wnd)
 
 void main_window_load_state(MainWindow *wnd)
 {
-	GSettings *settings = g_settings_new(CONFIG_WIN_STATE);
+	if(!gtk_widget_get_realized(GTK_WIDGET(wnd)))
+	{
+		GSettings *settings = g_settings_new(CONFIG_WIN_STATE);
+		gint width = g_settings_get_int(settings, "width");
+		gint height = g_settings_get_int(settings, "height");
+		gint handle_pos;
+		gdouble volume;
 
-	wnd->init_width = g_settings_get_int(settings, "width");
-	wnd->init_height = g_settings_get_int(settings, "height");
+		wnd->playlist_width
+			= g_settings_get_int(settings, "playlist-width");
+		wnd->playlist_visible
+			= g_settings_get_boolean(settings, "show-playlist");
+		volume = g_settings_get_double(settings, "volume");
+		handle_pos = width-(wnd->playlist_visible?wnd->playlist_width:0);
 
-	wnd->init_playlist_visible
-		= g_settings_get_boolean(settings, "show-playlist");
+		control_box_set_volume(CONTROL_BOX(wnd->control_box), volume);
+		gtk_widget_set_visible(wnd->playlist, wnd->playlist_visible);
+		gtk_window_resize(GTK_WINDOW(wnd), width, height);
+		gtk_paned_set_position(	GTK_PANED(wnd->vid_area_paned),
+					 handle_pos );
 
-	g_signal_connect(	wnd,
-				"configure-event",
-				G_CALLBACK(configure_handler),
-				wnd );
-
-	/* This is needed even when show_playlist==false to initialize some
-	 * internal variables.
-	 */
-	main_window_set_playlist_visible(wnd, TRUE);
-
-	wnd->playlist_width = g_settings_get_int(settings, "playlist-width");
-
-	control_box_set_volume(	CONTROL_BOX(wnd->control_box),
-				g_settings_get_double(settings, "volume") );
-
-	gtk_window_resize(GTK_WINDOW(wnd), wnd->init_width, wnd->init_height);
-
-	g_clear_object(&settings);
+		g_clear_object(&settings);
+	}
+	else
+	{
+		g_critical(	"Attempted to call main_window_load_state() on "
+				"realized window" );
+	}
 }
 
 void main_window_update_track_list(	MainWindow *wnd,
@@ -340,243 +611,24 @@ void main_window_update_track_list(	MainWindow *wnd,
 	}
 }
 
-static void main_window_class_init(MainWindowClass *klass)
+void main_window_resize_video_area(	MainWindow *wnd,
+					gint width,
+					gint height )
 {
-	GObjectClass *obj_class = G_OBJECT_CLASS(klass);
-	GtkWidgetClass *wgt_class = GTK_WIDGET_CLASS(klass);
-	GParamSpec *pspec = NULL;
-
-	obj_class->set_property = main_window_set_property;
-	obj_class->get_property = main_window_get_property;
-	wgt_class->motion_notify_event = motion_notify_handler;
-
-	g_signal_new(	"mpv-init",
-			G_TYPE_FROM_CLASS(klass),
-			G_SIGNAL_RUN_FIRST,
-			0,
-			NULL,
-			NULL,
-			g_cclosure_marshal_VOID__VOID,
-			G_TYPE_NONE,
-			0 );
-
-	g_signal_new(	"mpv-playback-restart",
-			G_TYPE_FROM_CLASS(klass),
-			G_SIGNAL_RUN_FIRST,
-			0,
-			NULL,
-			NULL,
-			g_cclosure_marshal_VOID__VOID,
-			G_TYPE_NONE,
-			0 );
-
-	g_signal_new(	"mpv-prop-change",
-			G_TYPE_FROM_CLASS(klass),
-			G_SIGNAL_RUN_FIRST,
-			0,
-			NULL,
-			NULL,
-			g_cclosure_marshal_VOID__STRING,
-			G_TYPE_NONE,
-			1,
-			G_TYPE_STRING );
-
-	pspec = g_param_spec_boolean
-		(	"use-opengl",
-			"Use OpenGL",
-			"Whether or not to set up video area for opengl-cb",
-			FALSE,
-			G_PARAM_CONSTRUCT_ONLY|G_PARAM_READWRITE );
-
-	g_object_class_install_property(obj_class, PROP_USE_OPENGL, pspec);
-}
-
-static void main_window_init(MainWindow *wnd)
-{
-	/* wnd->vid_area will be initialized when use-opengl property is set */
-	wnd->priv = main_window_get_instance_private(wnd);
-	wnd->fullscreen = FALSE;
-	wnd->playlist_visible = FALSE;
-	wnd->fs_control_hover = FALSE;
-	wnd->playlist_width = PLAYLIST_DEFAULT_WIDTH;
-	wnd->timeout_tag = 0;
-	wnd->init_width = MAIN_WINDOW_DEFAULT_WIDTH;
-	wnd->init_height = MAIN_WINDOW_DEFAULT_HEIGHT;
-	wnd->settings = gtk_settings_get_default();
-	wnd->header_bar = gtk_header_bar_new();
-	wnd->open_hdr_btn = NULL;
-	wnd->fullscreen_hdr_btn = NULL;
-	wnd->menu_hdr_btn = NULL;
-	wnd->main_box = gtk_box_new(GTK_ORIENTATION_VERTICAL, 0);
-	wnd->vid_area_paned = gtk_paned_new(GTK_ORIENTATION_HORIZONTAL);
-	wnd->vid_area_overlay = gtk_overlay_new();
-	wnd->control_box = control_box_new();
-	wnd->playlist = playlist_widget_new();
-
-	gtk_widget_add_events(	wnd->vid_area_overlay,
-				GDK_ENTER_NOTIFY_MASK
-				|GDK_LEAVE_NOTIFY_MASK );
-
-	gtk_header_bar_set_show_close_button(	GTK_HEADER_BAR(wnd->header_bar),
-						TRUE );
-
-	gtk_window_set_title(GTK_WINDOW(wnd), g_get_application_name());
-
-	gtk_paned_set_position(	GTK_PANED(wnd->vid_area_paned),
-				MAIN_WINDOW_DEFAULT_WIDTH
-				-PLAYLIST_DEFAULT_WIDTH );
-
-	gtk_window_set_default_size(	GTK_WINDOW(wnd),
-					MAIN_WINDOW_DEFAULT_WIDTH,
-					MAIN_WINDOW_DEFAULT_HEIGHT );
-
-	gtk_box_pack_start
-		(GTK_BOX(wnd->main_box), wnd->vid_area_paned, TRUE, TRUE, 0);
-
-	gtk_paned_pack2
-		(GTK_PANED(wnd->vid_area_paned), wnd->playlist, FALSE, TRUE);
-
-	gtk_container_add
-		(GTK_CONTAINER(wnd->main_box), wnd->control_box);
-
-	gtk_container_add
-		(GTK_CONTAINER(wnd), wnd->main_box);
-
-	g_signal_connect(	wnd->vid_area_overlay,
-				"enter-notify-event",
-				G_CALLBACK(fs_control_enter_handler),
+	g_signal_connect(	wnd->vid_area,
+				"size-allocate",
+				G_CALLBACK(size_allocate_handler),
 				wnd );
 
-	g_signal_connect(	wnd->vid_area_overlay,
-				"leave-notify-event",
-				G_CALLBACK(fs_control_leave_handler),
-				wnd );
-}
+	wnd->priv->resize_target[0] = width;
+	wnd->priv->resize_target[1] = height;
+	resize_to_target(wnd);
 
-GtkWidget *main_window_new(GtkApplication *app, gboolean use_opengl)
-{
-	return GTK_WIDGET(g_object_new(	main_window_get_type(),
-					"application", app,
-					"use-opengl", use_opengl,
-					NULL ));
-}
-
-void main_window_toggle_fullscreen(MainWindow *wnd)
-{
-	ControlBox *control_box = CONTROL_BOX(wnd->control_box);
-	GtkContainer* main_box = GTK_CONTAINER(wnd->main_box);
-	GtkContainer *overlay = GTK_CONTAINER(wnd->vid_area_overlay);
-
-	if(wnd->fullscreen)
-	{
-		GdkCursor *cursor;
-		GdkWindow *vid_area;
-
-		gtk_widget_set_halign(wnd->control_box, GTK_ALIGN_FILL);
-		gtk_widget_set_valign(wnd->control_box, GTK_ALIGN_FILL);
-		gtk_widget_set_size_request(wnd->control_box, -1, -1);
-
-		g_object_ref(wnd->control_box);
-		gtk_container_remove(overlay, wnd->control_box);
-		gtk_container_add(main_box, wnd->control_box);
-		g_object_unref(wnd->control_box);
-
-		control_box_set_fullscreen_state(control_box, FALSE);
-		gtk_window_unfullscreen(GTK_WINDOW(wnd));
-		gtk_widget_show(wnd->control_box);
-
-		if(main_window_get_csd_enabled(wnd))
-		{
-			control_box_set_fullscreen_btn_visible
-				(CONTROL_BOX(wnd->control_box), FALSE);
-		}
-		else
-		{
-			gtk_application_window_set_show_menubar
-				(GTK_APPLICATION_WINDOW(wnd), TRUE);
-		}
-
-		if(wnd->playlist_visible)
-		{
-			gtk_widget_show(wnd->playlist);
-		}
-
-		cursor =	gdk_cursor_new_from_name
-				(gdk_display_get_default(), "default");
-
-		vid_area = gtk_widget_get_window(GTK_WIDGET(wnd->vid_area));
-
-		gdk_window_set_cursor(vid_area, cursor);
-
-		wnd->fullscreen = FALSE;
-	}
-	else
-	{
-		GdkScreen *screen;
-		GdkWindow *window;
-		GdkRectangle monitor_geom;
-		gint width;
-		gint monitor;
-
-		screen = gtk_window_get_screen(GTK_WINDOW(wnd));
-		window = gtk_widget_get_window(GTK_WIDGET(wnd));
-		monitor = gdk_screen_get_monitor_at_window(screen, window);
-
-		gdk_screen_get_monitor_geometry(screen, monitor, &monitor_geom);
-
-		width = monitor_geom.width/2;
-
-		gtk_widget_set_halign(wnd->control_box, GTK_ALIGN_CENTER);
-		gtk_widget_set_valign(wnd->control_box, GTK_ALIGN_END);
-		gtk_widget_set_size_request(wnd->control_box, width, -1);
-
-		g_object_ref(wnd->control_box);
-		gtk_container_remove(main_box, wnd->control_box);
-		gtk_overlay_add_overlay(GTK_OVERLAY(overlay), wnd->control_box);
-		g_object_unref(wnd->control_box);
-
-		control_box_set_fullscreen_state(control_box, TRUE);
-		gtk_window_fullscreen(GTK_WINDOW(wnd));
-		gtk_window_present(GTK_WINDOW(wnd));
-		gtk_widget_hide(wnd->control_box);
-
-		if(main_window_get_csd_enabled(wnd))
-		{
-			control_box_set_fullscreen_btn_visible
-				(CONTROL_BOX(wnd->control_box), TRUE);
-		}
-		else
-		{
-			gtk_application_window_set_show_menubar
-				(GTK_APPLICATION_WINDOW(wnd), FALSE);
-		}
-
-		if(wnd->playlist_visible)
-		{
-			gtk_widget_hide(wnd->playlist);
-		}
-
-		wnd->fullscreen = TRUE;
-		timeout_handler(wnd);
-	}
-}
-
-void main_window_reset(MainWindow *wnd)
-{
-	gtk_window_set_title(GTK_WINDOW(wnd), g_get_application_name());
-	control_box_reset_control(CONTROL_BOX(wnd->control_box));
-}
-
-gint main_window_get_width_margin(MainWindow *wnd)
-{
-	return	gtk_widget_get_allocated_width(GTK_WIDGET(wnd))
-		- gtk_widget_get_allocated_width(wnd->vid_area);
-}
-
-gint main_window_get_height_margin(MainWindow *wnd)
-{
-	return	gtk_widget_get_allocated_height(GTK_WIDGET(wnd))
-		- gtk_widget_get_allocated_height(wnd->vid_area);
+	/* The size may not change, so this is needed to ensure that
+	 * size_allocate_handler() will be called so that the event handler will
+	 * be disconnected.
+	 */
+	gtk_widget_queue_allocate(wnd->vid_area);
 }
 
 gboolean main_window_get_use_opengl(MainWindow *wnd)
@@ -597,14 +649,12 @@ void main_window_enable_csd(MainWindow *wnd)
 
 	open_icon = g_themed_icon_new_with_default_fallbacks
 				("list-add-symbolic");
-
 	fullscreen_icon = g_themed_icon_new_with_default_fallbacks
 				("view-fullscreen-symbolic");
-
 	menu_icon = g_themed_icon_new_with_default_fallbacks
 				("view-list-symbolic");
 
-	wnd->playlist_width = PLAYLIST_DEFAULT_WIDTH+PLAYLIST_CSD_WIDTH_OFFSET;
+	wnd->playlist_width = PLAYLIST_DEFAULT_WIDTH;
 	wnd->open_hdr_btn = gtk_menu_button_new();
 	wnd->fullscreen_hdr_btn = gtk_button_new();
 	wnd->menu_hdr_btn = gtk_menu_button_new();
@@ -620,41 +670,38 @@ void main_window_enable_csd(MainWindow *wnd)
 		(	GTK_BUTTON(wnd->fullscreen_hdr_btn),
 			gtk_image_new_from_gicon
 				(fullscreen_icon, GTK_ICON_SIZE_MENU ));
-
 	gtk_button_set_image
 		(	GTK_BUTTON(wnd->open_hdr_btn),
 			gtk_image_new_from_gicon
 				(open_icon, GTK_ICON_SIZE_MENU ));
-
 	gtk_button_set_image
 		(	GTK_BUTTON(wnd->menu_hdr_btn),
 			gtk_image_new_from_gicon
 				(menu_icon, GTK_ICON_SIZE_MENU ));
-
 	gtk_menu_button_set_menu_model
 		(	GTK_MENU_BUTTON(wnd->open_hdr_btn),
 			G_MENU_MODEL(open_btn_menu) );
-
 	gtk_menu_button_set_menu_model
 		(	GTK_MENU_BUTTON(wnd->menu_hdr_btn),
 			G_MENU_MODEL(menu_btn_menu) );
 
 	gtk_header_bar_pack_start
 		(GTK_HEADER_BAR(wnd->header_bar), wnd->open_hdr_btn);
-
 	gtk_header_bar_pack_end
 		(GTK_HEADER_BAR(wnd->header_bar), wnd->menu_hdr_btn);
-
 	gtk_header_bar_pack_end
 		(GTK_HEADER_BAR(wnd->header_bar), wnd->fullscreen_hdr_btn);
 
+	gtk_widget_set_tooltip_text(	wnd->fullscreen_hdr_btn,
+					_("Toggle Fullscreen") );
+
 	gtk_actionable_set_action_name
-		(GTK_ACTIONABLE(wnd->fullscreen_hdr_btn), "app.fullscreen");
+		(	GTK_ACTIONABLE(wnd->fullscreen_hdr_btn),
+			"app.fullscreen_toggle" );
 
 	gtk_paned_set_position(	GTK_PANED(wnd->vid_area_paned),
 				MAIN_WINDOW_DEFAULT_WIDTH
-				-PLAYLIST_DEFAULT_WIDTH
-				-PLAYLIST_CSD_WIDTH_OFFSET );
+				-PLAYLIST_DEFAULT_WIDTH );
 
 	gtk_window_set_titlebar(GTK_WINDOW(wnd), wnd->header_bar);
 	gtk_window_set_title(GTK_WINDOW(wnd), g_get_application_name());
@@ -669,33 +716,46 @@ gboolean main_window_get_csd_enabled(MainWindow *wnd)
 
 void main_window_set_playlist_visible(MainWindow *wnd, gboolean visible)
 {
-	gint offset;
-	gint handle_pos;
-	gint width;
-	gint height;
-
-	offset = main_window_get_csd_enabled(wnd)?PLAYLIST_CSD_WIDTH_OFFSET:0;
-	handle_pos = gtk_paned_get_position(GTK_PANED(wnd->vid_area_paned));
-
-	gtk_window_get_size(GTK_WINDOW(wnd), &width, &height);
-
-	if(!visible && wnd->playlist_visible)
+	if(visible != wnd->playlist_visible && !wnd->fullscreen)
 	{
-		wnd->playlist_width = width-handle_pos;
+		gboolean maximized;
+		gint handle_pos;
+		gint width;
+		gint height;
+
+		maximized = gtk_window_is_maximized(GTK_WINDOW(wnd));
+		handle_pos =	gtk_paned_get_position
+				(GTK_PANED(wnd->vid_area_paned));
+
+		gtk_window_get_size(GTK_WINDOW(wnd), &width, &height);
+
+		if(wnd->priv->playlist_first_toggle && visible)
+		{
+			gint new_pos = width-(maximized?wnd->playlist_width:0);
+
+			gtk_paned_set_position(	GTK_PANED(wnd->vid_area_paned),
+						new_pos );
+		}
+		else if(!visible)
+		{
+			wnd->playlist_width = width-handle_pos;
+		}
+
+
+		wnd->playlist_visible = visible;
+		gtk_widget_set_visible(wnd->playlist, visible);
+
+		if(!maximized)
+		{
+			gtk_window_resize(	GTK_WINDOW(wnd),
+						visible?
+						width+wnd->playlist_width:
+						handle_pos,
+						height );
+		}
+
+		wnd->priv->playlist_first_toggle = FALSE;
 	}
-
-	wnd->playlist_visible = visible;
-
-	gtk_widget_set_visible(wnd->playlist, visible);
-
-	/* For some unknown reason, width needs to be adjusted by some offset
-	 * (50px) when CSD is enabled for the resulting size to be correct.
-	 */
-	gtk_window_resize(	GTK_WINDOW(wnd),
-				visible
-				?width+wnd->playlist_width-offset
-				:handle_pos+offset,
-				height );
 }
 
 gboolean main_window_get_playlist_visible(MainWindow *wnd)
