@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2015 gnome-mpv
+ * Copyright (c) 2015-2016 gnome-mpv
  *
  * This file is part of GNOME MPV.
  *
@@ -21,11 +21,11 @@
 
 #include "mpris_player.h"
 #include "mpris_gdbus.h"
-#include "mpv.h"
+#include "mpv_obj.h"
 #include "def.h"
 
-static void set_paused(gmpv_handle *ctx, gboolean paused);
 static void prop_table_init(mpris *inst);
+static void emit_prop_changed(mpris *inst, const mpris_prop_val_pair *prop_list);
 static void method_handler(	GDBusConnection *connection,
 				const gchar *sender,
 				const gchar *object_path,
@@ -60,16 +60,6 @@ static void mpv_prop_change_handler(	MainWindow *wnd,
 					gchar *name,
 					gpointer data );
 
-static void set_paused(gmpv_handle *ctx, gboolean paused)
-{
-	ctx->paused = paused;
-
-	mpv_set_property(	ctx->mpv_ctx,
-				"pause",
-				MPV_FORMAT_FLAG,
-				&ctx->paused );
-}
-
 static void prop_table_init(mpris *inst)
 {
 	/* Position is retrieved from mpv on-demand */
@@ -98,6 +88,14 @@ static void prop_table_init(mpris *inst)
 	}
 }
 
+static void emit_prop_changed(mpris *inst, const mpris_prop_val_pair *prop_list)
+{
+	GDBusInterfaceInfo *iface;
+
+	iface = mpris_org_mpris_media_player2_player_interface_info();
+	mpris_emit_prop_changed(inst, iface->name, prop_list);
+}
+
 static void method_handler(	GDBusConnection *connection,
 				const gchar *sender,
 				const gchar *object_path,
@@ -113,31 +111,37 @@ static void method_handler(	GDBusConnection *connection,
 	{
 		const gchar *cmd[] = {"playlist_next", "weak", NULL};
 
-		mpv_command(inst->gmpv_ctx->mpv_ctx, cmd);
+		mpv_command(inst->gmpv_ctx->mpv->mpv_ctx, cmd);
 	}
 	else if(g_strcmp0(method_name, "Previous") == 0)
 	{
 		const gchar *cmd[] = {"playlist_prev", "weak", NULL};
 
-		mpv_command(inst->gmpv_ctx->mpv_ctx, cmd);
+		mpv_command(inst->gmpv_ctx->mpv->mpv_ctx, cmd);
 	}
 	else if(g_strcmp0(method_name, "Pause") == 0)
 	{
-		set_paused(inst->gmpv_ctx, TRUE);
+		mpv_obj_set_property_flag(inst->gmpv_ctx->mpv, "pause", TRUE);
 	}
 	else if(g_strcmp0(method_name, "PlayPause") == 0)
 	{
-		set_paused(inst->gmpv_ctx, !inst->gmpv_ctx->paused);
+		gboolean paused;
+
+		paused =	mpv_obj_get_property_flag
+				(inst->gmpv_ctx->mpv, "pause");
+
+		mpv_obj_set_property_flag
+			(inst->gmpv_ctx->mpv, "pause", !paused);
 	}
 	else if(g_strcmp0(method_name, "Stop") == 0)
 	{
 		const gchar *cmd[] = {"stop", NULL};
 
-		mpv_command(inst->gmpv_ctx->mpv_ctx, cmd);
+		mpv_command(inst->gmpv_ctx->mpv->mpv_ctx, cmd);
 	}
 	else if(g_strcmp0(method_name, "Play") == 0)
 	{
-		set_paused(inst->gmpv_ctx, FALSE);
+		mpv_obj_set_property_flag(inst->gmpv_ctx->mpv, "pause", FALSE);
 	}
 	else if(g_strcmp0(method_name, "Seek") == 0)
 	{
@@ -146,14 +150,14 @@ static void method_handler(	GDBusConnection *connection,
 
 		g_variant_get(parameters, "(x)", &offset_us);
 
-		mpv_get_property(	inst->gmpv_ctx->mpv_ctx,
+		mpv_get_property(	inst->gmpv_ctx->mpv->mpv_ctx,
 					"time-pos",
 					MPV_FORMAT_DOUBLE,
 					&position );
 
 		position += (gdouble)offset_us/1.0e6;
 
-		mpv_set_property(	inst->gmpv_ctx->mpv_ctx,
+		mpv_set_property(	inst->gmpv_ctx->mpv->mpv_ctx,
 					"time-pos",
 					MPV_FORMAT_DOUBLE,
 					&position );
@@ -183,14 +187,14 @@ static void method_handler(	GDBusConnection *connection,
 			inst->pending_seek = (gdouble)time_us/1.0e6;
 			index = g_ascii_strtoll(track_id+prefix_len, NULL, 0);
 
-			mpv_get_property(	inst->gmpv_ctx->mpv_ctx,
+			mpv_get_property(	inst->gmpv_ctx->mpv->mpv_ctx,
 						"playlist-pos",
 						MPV_FORMAT_INT64,
 						&old_index );
 
 			if(index != old_index)
 			{
-				mpv_set_property(	inst->gmpv_ctx->mpv_ctx,
+				mpv_set_property(	inst->gmpv_ctx->mpv->mpv_ctx,
 							"playlist-pos",
 							MPV_FORMAT_INT64,
 							&index );
@@ -207,7 +211,7 @@ static void method_handler(	GDBusConnection *connection,
 		const gchar *uri;
 
 		g_variant_get(parameters, "(&s)", &uri);
-		mpv_load(inst->gmpv_ctx, uri, FALSE, TRUE);
+		mpv_obj_load(inst->gmpv_ctx->mpv, uri, FALSE, TRUE);
 	}
 
 	g_dbus_method_invocation_return_value
@@ -230,7 +234,7 @@ static GVariant *get_prop_handler(	GDBusConnection *connection,
 		gdouble position;
 		gint rc;
 
-		rc = mpv_get_property(	inst->gmpv_ctx->mpv_ctx,
+		rc = mpv_get_property(	inst->gmpv_ctx->mpv->mpv_ctx,
 					"time-pos",
 					MPV_FORMAT_DOUBLE,
 					&position );
@@ -261,7 +265,7 @@ static gboolean set_prop_handler(	GDBusConnection *connection,
 	{
 		gdouble rate = g_variant_get_double(value);
 
-		mpv_set_property(	inst->gmpv_ctx->mpv_ctx,
+		mpv_set_property(	inst->gmpv_ctx->mpv->mpv_ctx,
 					"speed",
 					MPV_FORMAT_DOUBLE,
 					&rate );
@@ -270,7 +274,7 @@ static gboolean set_prop_handler(	GDBusConnection *connection,
 	{
 		gdouble volume = 100*g_variant_get_double(value);
 
-		mpv_set_property(	inst->gmpv_ctx->mpv_ctx,
+		mpv_set_property(	inst->gmpv_ctx->mpv->mpv_ctx,
 					"volume",
 					MPV_FORMAT_DOUBLE,
 					&volume );
@@ -293,12 +297,12 @@ static void playback_status_update_handler(mpris *inst)
 	gint core_idle;
 	gboolean can_seek;
 
-	mpv_get_property(	inst->gmpv_ctx->mpv_ctx,
+	mpv_get_property(	inst->gmpv_ctx->mpv->mpv_ctx,
 				"idle",
 				MPV_FORMAT_FLAG,
 				&idle );
 
-	mpv_get_property(	inst->gmpv_ctx->mpv_ctx,
+	mpv_get_property(	inst->gmpv_ctx->mpv->mpv_ctx,
 				"core-idle",
 				MPV_FORMAT_FLAG,
 				&core_idle );
@@ -337,7 +341,7 @@ static void playback_status_update_handler(mpris *inst)
 				{"CanSeek", can_seek_value},
 				{NULL, NULL} };
 
-	mpris_emit_prop_changed(inst, prop_list);
+	emit_prop_changed(inst, prop_list);
 }
 
 static void playlist_update_handler(mpris *inst)
@@ -351,12 +355,12 @@ static void playlist_update_handler(mpris *inst)
 	gint64 playlist_pos;
 	gint rc = 0;
 
-	rc |= mpv_get_property(	inst->gmpv_ctx->mpv_ctx,
+	rc |= mpv_get_property(	inst->gmpv_ctx->mpv->mpv_ctx,
 				"playlist-count",
 				MPV_FORMAT_INT64,
 				&playlist_count );
 
-	rc |= mpv_get_property(	inst->gmpv_ctx->mpv_ctx,
+	rc |= mpv_get_property(	inst->gmpv_ctx->mpv->mpv_ctx,
 				"playlist-pos",
 				MPV_FORMAT_INT64,
 				&playlist_pos );
@@ -379,7 +383,7 @@ static void playlist_update_handler(mpris *inst)
 				{"CanGoNext", can_next_value},
 				{NULL, NULL} };
 
-	mpris_emit_prop_changed(inst, prop_list);
+	emit_prop_changed(inst, prop_list);
 }
 
 static void speed_update_handler(mpris *inst)
@@ -388,7 +392,7 @@ static void speed_update_handler(mpris *inst)
 	GVariant *value;
 	mpris_prop_val_pair *prop_list;
 
-	mpv_get_property(	inst->gmpv_ctx->mpv_ctx,
+	mpv_get_property(	inst->gmpv_ctx->mpv->mpv_ctx,
 				"speed",
 				MPV_FORMAT_DOUBLE,
 				&speed );
@@ -402,7 +406,7 @@ static void speed_update_handler(mpris *inst)
 	prop_list =	(mpris_prop_val_pair[])
 			{{"Rate", value}, {NULL, NULL}};
 
-	mpris_emit_prop_changed(inst, prop_list);
+	emit_prop_changed(inst, prop_list);
 }
 
 static void metadata_update_handler(mpris *inst)
@@ -426,6 +430,8 @@ static void metadata_update_handler(mpris *inst)
 	GVariantBuilder builder;
 	GVariant *value;
 	gchar *uri;
+	gchar *playlist_pos_str;
+	gchar *trackid;
 	gdouble duration;
 	gint64 playlist_pos;
 	gint rc;
@@ -433,7 +439,7 @@ static void metadata_update_handler(mpris *inst)
 
 	g_variant_builder_init(&builder, G_VARIANT_TYPE("a{sv}"));
 
-	uri = mpv_get_property_string(inst->gmpv_ctx->mpv_ctx, "path");
+	uri = mpv_get_property_string(inst->gmpv_ctx->mpv->mpv_ctx, "path");
 
 	if(uri)
 	{
@@ -445,49 +451,38 @@ static void metadata_update_handler(mpris *inst)
 		mpv_free(uri);
 	}
 
-	rc = mpv_get_property(	inst->gmpv_ctx->mpv_ctx,
+	rc = mpv_get_property(	inst->gmpv_ctx->mpv->mpv_ctx,
 				"duration",
 				MPV_FORMAT_DOUBLE,
 				&duration );
 
-	if(rc >= 0)
-	{
-		gint64 length = (gint64)(duration*1e6);
+	g_variant_builder_add(	&builder,
+				"{sv}",
+				"mpris:length",
+				g_variant_new_int64
+				((gint64)((rc >= 0)*duration*1e6)) );
 
-		g_variant_builder_add(	&builder,
-					"{sv}",
-					"mpris:length",
-					g_variant_new_int64(length) );
-	}
-
-	rc = mpv_get_property(	inst->gmpv_ctx->mpv_ctx,
+	rc = mpv_get_property(	inst->gmpv_ctx->mpv->mpv_ctx,
 				"playlist-pos",
 				MPV_FORMAT_INT64,
 				&playlist_pos );
 
-	if(rc >= 0)
-	{
-		gchar *playlist_pos_str;
-		gchar *trackid;
+	playlist_pos_str = g_strdup_printf(	"%" G_GINT64_FORMAT,
+						(rc >= 0)*playlist_pos );
 
-		playlist_pos_str = g_strdup_printf(	"%" G_GINT64_FORMAT,
-							playlist_pos );
+	trackid = g_strconcat(	MPRIS_TRACK_ID_PREFIX,
+				playlist_pos_str,
+				NULL );
 
-		trackid = g_strconcat(	MPRIS_TRACK_ID_PREFIX,
-					playlist_pos_str,
-					NULL );
-
-		g_variant_builder_add(	&builder,
-					"{sv}",
-					"mpris:trackid",
-					g_variant_new_object_path(trackid) );
-
-		g_free(playlist_pos_str);
-		g_free(trackid);
-	}
+	g_variant_builder_add(	&builder,
+				"{sv}",
+				"mpris:trackid",
+				g_variant_new_object_path(trackid) );
 
 	for(i = 0; tag_map[i].mpv_name; i++)
 	{
+		GVariantBuilder tag_builder;
+		GVariant *tag_value;
 		gchar *mpv_prop;
 		gchar *value_buf;
 
@@ -496,14 +491,11 @@ static void metadata_update_handler(mpris *inst)
 					NULL );
 
 		value_buf = mpv_get_property_string
-				(inst->gmpv_ctx->mpv_ctx, mpv_prop);
+				(inst->gmpv_ctx->mpv->mpv_ctx, mpv_prop);
 
 		if(value_buf)
 		{
-			GVariantBuilder tag_builder;
-			GVariant *tag_value;
-
-			tag_value = g_variant_new_string(value_buf);
+			tag_value = g_variant_new_string(value_buf?:"");
 
 			if(tag_map[i].is_array)
 			{
@@ -521,9 +513,10 @@ static void metadata_update_handler(mpris *inst)
 					tag_map[i].is_array?
 					g_variant_new("as", &tag_builder):
 					tag_value );
+
+			mpv_free(value_buf);
 		}
 
-		mpv_free(value_buf);
 		g_free(mpv_prop);
 	}
 
@@ -536,8 +529,11 @@ static void metadata_update_handler(mpris *inst)
 	prop_list =	(mpris_prop_val_pair[])
 			{{"Metadata", value}, {NULL, NULL}};
 
-	mpris_emit_prop_changed(inst, prop_list);
+	emit_prop_changed(inst, prop_list);
 	mpv_playback_restart_handler(inst->gmpv_ctx->gui, inst);
+
+	g_free(playlist_pos_str);
+	g_free(trackid);
 }
 
 static void volume_update_handler(mpris *inst)
@@ -546,58 +542,58 @@ static void volume_update_handler(mpris *inst)
 	gdouble volume;
 	GVariant *value;
 
-	mpv_get_property(	inst->gmpv_ctx->mpv_ctx,
+	mpv_get_property(	inst->gmpv_ctx->mpv->mpv_ctx,
 				"volume",
 				MPV_FORMAT_DOUBLE,
 				&volume );
 
 	value = g_variant_new_double(volume/100.0);
 
-	g_hash_table_replace (	inst->player_prop_table,
+	g_hash_table_replace(	inst->player_prop_table,
 				"Volume",
 				g_variant_ref(value) );
 
 	prop_list =	(mpris_prop_val_pair[])
 			{{"Volume", value}, {NULL, NULL}};
 
-	mpris_emit_prop_changed(inst, prop_list);
+	emit_prop_changed(inst, prop_list);
 }
 
 static void mpv_init_handler(MainWindow *wnd, gpointer data)
 {
 	mpris *inst = data;
 
-	mpv_observe_property(	inst->gmpv_ctx->mpv_ctx,
+	mpv_observe_property(	inst->gmpv_ctx->mpv->mpv_ctx,
 				0,
 				"idle",
 				MPV_FORMAT_FLAG );
 
-	mpv_observe_property(	inst->gmpv_ctx->mpv_ctx,
+	mpv_observe_property(	inst->gmpv_ctx->mpv->mpv_ctx,
 				0,
 				"core-idle",
 				MPV_FORMAT_FLAG );
 
-	mpv_observe_property(	inst->gmpv_ctx->mpv_ctx,
+	mpv_observe_property(	inst->gmpv_ctx->mpv->mpv_ctx,
 				0,
 				"speed",
 				MPV_FORMAT_DOUBLE );
 
-	mpv_observe_property(	inst->gmpv_ctx->mpv_ctx,
+	mpv_observe_property(	inst->gmpv_ctx->mpv->mpv_ctx,
 				0,
 				"metadata",
 				MPV_FORMAT_NODE );
 
-	mpv_observe_property(	inst->gmpv_ctx->mpv_ctx,
+	mpv_observe_property(	inst->gmpv_ctx->mpv->mpv_ctx,
 				0,
 				"volume",
 				MPV_FORMAT_DOUBLE );
 
-	mpv_observe_property(	inst->gmpv_ctx->mpv_ctx,
+	mpv_observe_property(	inst->gmpv_ctx->mpv->mpv_ctx,
 				0,
 				"playlist-pos",
 				MPV_FORMAT_INT64 );
 
-	mpv_observe_property(	inst->gmpv_ctx->mpv_ctx,
+	mpv_observe_property(	inst->gmpv_ctx->mpv->mpv_ctx,
 				0,
 				"playlist-count",
 				MPV_FORMAT_INT64 );
@@ -614,14 +610,14 @@ static void mpv_playback_restart_handler(MainWindow *wnd, gpointer data)
 		position = inst->pending_seek;
 		inst->pending_seek = -1;
 
-		mpv_set_property(	inst->gmpv_ctx->mpv_ctx,
+		mpv_set_property(	inst->gmpv_ctx->mpv->mpv_ctx,
 					"time-pos",
 					MPV_FORMAT_DOUBLE,
 					&position );
 	}
 	else
 	{
-		mpv_get_property(	inst->gmpv_ctx->mpv_ctx,
+		mpv_get_property(	inst->gmpv_ctx->mpv->mpv_ctx,
 					"time-pos",
 					MPV_FORMAT_DOUBLE,
 					&position );
@@ -686,19 +682,19 @@ void mpris_player_register(mpris *inst)
 	inst->player_sig_id_list = g_malloc(4*sizeof(gulong));
 
 	inst->player_sig_id_list[0]
-		= g_signal_connect(	inst->gmpv_ctx->gui,
+		= g_signal_connect(	inst->gmpv_ctx->mpv,
 					"mpv-init",
 					G_CALLBACK(mpv_init_handler),
 					inst );
 
 	inst->player_sig_id_list[1]
-		= g_signal_connect(	inst->gmpv_ctx->gui,
+		= g_signal_connect(	inst->gmpv_ctx->mpv,
 					"mpv-playback-restart",
 					G_CALLBACK(mpv_playback_restart_handler),
 					inst );
 
 	inst->player_sig_id_list[2]
-		= g_signal_connect(	inst->gmpv_ctx->gui,
+		= g_signal_connect(	inst->gmpv_ctx->mpv,
 					"mpv-prop-change",
 					G_CALLBACK(mpv_prop_change_handler),
 					inst );
