@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014 gnome-mpv
+ * Copyright (c) 2014-2016 gnome-mpv
  *
  * This file is part of GNOME MPV.
  *
@@ -22,6 +22,8 @@
 
 #include <gtk/gtk.h>
 
+#include "playlist.h"
+
 G_BEGIN_DECLS
 
 #define PLAYLIST_WIDGET_TYPE (playlist_widget_get_type ())
@@ -40,22 +42,12 @@ G_BEGIN_DECLS
 #define	IS_PLAYLIST_WIDGET_CLASS(klass) \
 	(G_TYPE_CHECK_CLASS_TYPE((klass), PLAYLIST_WIDGET_TYPE))
 
-enum PlaylistStoreColumn
-{
-	PLAYLIST_INDICATOR_COLUMN,
-	PLAYLIST_NAME_COLUMN,
-	PLAYLIST_URI_COLUMN,
-	PLAYLIST_N_COLUMNS
-};
-
 struct _PlaylistWidget
 {
 	GtkScrolledWindow parent_instance;
 	GtkWidget *tree_view;
-	GtkListStore *list_store;
-	GtkTreeViewColumn *indicator_column;
+	Playlist *store;
 	GtkTreeViewColumn *title_column;
-	GtkCellRenderer *indicator_renderer;
 	GtkCellRenderer *title_renderer;
 };
 
@@ -66,20 +58,10 @@ struct _PlaylistWidgetClass
 
 typedef struct _PlaylistWidget PlaylistWidget;
 typedef struct _PlaylistWidgetClass PlaylistWidgetClass;
-typedef enum PlaylistStoreColumn PlaylistStoreColumn;
 
-GtkWidget *playlist_widget_new(void);
+GtkWidget *playlist_widget_new(Playlist *store);
 GType playlist_widget_get_type(void);
-gboolean playlist_widget_empty(PlaylistWidget *wgt);
-void playlist_widget_clear(PlaylistWidget *wgt);
-void playlist_widget_get_iter_first(PlaylistWidget *wgt, GtkTreeIter *iter);
-void playlist_widget_iter_next(PlaylistWidget *wgt, GtkTreeIter *iter);
-void playlist_widget_get_uri(PlaylistWidget *wgt, gint pos);
-void playlist_widget_set_indicator_pos(PlaylistWidget *wgt, gint pos);
-void playlist_widget_remove(PlaylistWidget *wgt, gint pos);
-void playlist_widget_append(	PlaylistWidget *wgt,
-				const gchar *name,
-				const gchar *uri );
+void playlist_widget_remove_selected(PlaylistWidget *wgt);
 
 G_END_DECLS
 
