@@ -22,8 +22,6 @@
 
 #include <gtk/gtk.h>
 
-#include "pref_store.h"
-
 G_BEGIN_DECLS
 
 #define PREF_DIALOG_TYPE (pref_dialog_get_type ())
@@ -45,8 +43,6 @@ typedef struct _PrefDialogClass PrefDialogClass;
 
 GtkWidget *pref_dialog_new(GtkWindow *parent);
 GType pref_dialog_get_type(void);
-pref_store *pref_dialog_get_pref(PrefDialog *dlg);
-void pref_dialog_set_pref(PrefDialog *dlg, pref_store *pref);
 
 G_END_DECLS
 
